<?php

declare(strict_types=1);

namespace Drupal\graphql_compose_fragments;

use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\graphql_compose\Plugin\GraphQLComposeEntityTypeManager;
use Drupal\graphql_compose\Plugin\GraphQLComposeSchemaTypeManager;
use GraphQL\Type\Definition\HasFieldsType;
use GraphQL\Type\Definition\ObjectType;
use GraphQL\Type\Definition\ScalarType;
use GraphQL\Type\Definition\Type;
use GraphQL\Type\Definition\UnionType;
use GraphQL\Type\Definition\WrappingType;

/**
 * Controller for the GraphQL Compose fragment controller.
 */
class FragmentManager {

  use StringTranslationTrait;

  /**
   * Construct a new Fragment Controller.
   *
   * @param \Drupal\graphql_compose\Plugin\GraphQLComposeSchemaTypeManager $gqlSchemaTypeManager
   *   The GraphQL Compose schema type manager.
   * @param \Drupal\graphql_compose\Plugin\GraphQLComposeEntityTypeManager $gqlEntityTypeManager
   *   The GraphQL Compose entity type manager.
   */
  public function __construct(
    protected GraphQLComposeSchemaTypeManager $gqlSchemaTypeManager,
    protected GraphQLComposeEntityTypeManager $gqlEntityTypeManager,
  ) {}

  /**
   * Return the fragments.
   *
   * @return \GraphQL\Type\Definition\Type[]
   *   The schema types.
   */
  public function getTypes(): array {

    $types = [];

    // Load all entity types.
    foreach ($this->gqlEntityTypeManager->getPluginInstances() as $entity_type) {
      $entity_type->registerTypes();
    }

    // Load all types.
    foreach ($this->gqlSchemaTypeManager->getDefinitions() as $definition) {
      $this->gqlSchemaTypeManager->get($definition['id']);
    }

    // Expand definitions.
    foreach ($this->gqlSchemaTypeManager->getTypes() as $type) {
      if ($type instanceof HasFieldsType) {
        $type->getFields();
      }

      if ($type instanceof ObjectType) {
        $types[$type->name] = $type;
      }

      if ($type instanceof UnionType) {
        $type->getTypes();
        $types[$type->name] = $type;
      }
    }

    // Get the object extensions.
    $extensions = $this->getExtensions();

    // Extend the objects fields with extensions fields.
    foreach ($types as $type) {
      if ($type instanceof HasFieldsType && isset($extensions[$type->name])) {

        $types[$type->name] = new ObjectType([
          'name' => $type->name,
          'fields' => array_merge(
            $type->getFields(),
            $extensions[$type->name]->getFields(),
          ),
        ]);
      }
    }

    ksort($types);

    return $types;
  }

  /**
   * Get the object extensions.
   *
   * @return array
   *   The extensions.
   */
  public function getExtensions(): array {
    $extensions = [];

    foreach ($this->gqlSchemaTypeManager->getExtensions() as $extension) {
      if ($extension instanceof ObjectType) {
        $extensions[$extension->name] = $extension;
      }
    }

    return $extensions;
  }

  /**
   * Get the object fragment.
   *
   * @param \GraphQL\Type\Definition\Type $type
   *   The object type.
   *
   * @return array
   *   The object type and fragment.
   */
  public function getFragment(Type $type): array {
    $content = $this->getFragmentContent($type);
    $content = PHP_EOL . preg_replace('/^/m', '  ', $content) . PHP_EOL;

    return [
      'type' => $type,
      'name' => $this->t('Fragment@name', ['@name' => $type->name]),
      'content' => $this->t('fragment Fragment@name on @name {@content}', [
        '@name' => $type->name,
        '@content' => $content,
      ]),
    ];
  }

  /**
   * Get the object fields.
   *
   * @param \GraphQL\Type\Definition\Type $type
   *   The object type.
   *
   * @return string
   *   The object fields.
   */
  public function getFragmentContent(Type $type): string {
    $fields = [];

    if ($type instanceof UnionType) {
      foreach ($type->getTypes() as $union) {
        $fields[] = '  ' . $this->t('... Fragment@name', ['@name' => $union->name]);
      }
    }
    elseif ($type instanceof HasFieldsType) {
      foreach ($type->getFields() as $field) {

        // Unwrap the field type.
        $field_type = $field->getType();
        if ($field_type instanceof WrappingType) {
          $field_type = $field_type->getWrappedType(TRUE);
        }

        if (!$field_type instanceof ScalarType) {
          $content = ($field_type->name === $type->name)
            ? $this->t('# Recursion. Use best judgement or define manually')
            : $this->t('... Fragment@name', ['@name' => $field_type->name]);

          $fields[] = $field->name . ' {' . PHP_EOL . '  ' . $content . PHP_EOL . '}';
        }
        else {
          $fields[] = $field->name;
        }
      }
    }

    return implode(PHP_EOL, $fields);

  }

}
